@@ -1,19 +1,19 @@
 # -*- perl -*-
 
 #
-# $Id: Getopt.pm,v 1.48 2003/09/16 11:57:19 eserte Exp eserte $
+# $Id: Getopt.pm,v 1.65 2008/09/23 19:50:43 eserte Exp $
 # Author: Slaven Rezic
 #
-# Copyright (C) 1997,1998,1999,2000,2003 Slaven Rezic. All rights reserved.
+# Copyright (C) 1997,1998,1999,2000,2003,2007,2008 Slaven Rezic. All rights reserved.
 # This package is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 #
-# Mail: slaven@rezic.de
-# WWW:  http://user.cs.tu-berlin.de/~eserte/
+# Mail: srezic@cpan.org
+# WWW:  http://www.rezic.de/eserte/
 #
 
 package Tk::Getopt;
-require 5.003;
+require 5.005; # calling CODE refs
 use strict;
 use vars qw($loadoptions $VERSION $x11_pass_through
 	    $CHECKMARK_OFF $CHECKMARK_ON
@@ -24,7 +24,10 @@ use constant OPTTYPE  => 1;
 use constant DEFVAL   => 2;
 use constant OPTEXTRA => 3;
 
-$VERSION = '0.49';
+use Carp qw();
+
+$VERSION = '0.50';
+$VERSION = eval $VERSION;
 
 $DEBUG = 0;
 $x11_pass_through = 0;
@@ -130,9 +133,10 @@ sub new {
 	die "No opttable array ref or getopt hash ref";
     }
 
-    $self->{'caller'}   = (caller)[0];
-    $self->{'filename'} = delete $a{'-filename'};
-    $self->{'nosafe'}   = delete $a{'-nosafe'};
+    $self->{'caller'}         = (caller)[0];
+    $self->{'filename'}       = delete $a{'-filename'};
+    $self->{'nosafe'}         = delete $a{'-nosafe'};
+    $self->{'useerrordialog'} = delete $a{'-useerrordialog'};
 
     die "Unrecognized arguments: " . join(" ", %a) if %a;
 
@@ -153,7 +157,7 @@ sub _opt_array {
 }
 
 # Return a reference to the option variable given by $opt
-sub _varref {
+sub varref {
     my($self, $opt) = @_;
     if($opt->[OPTEXTRA]{'var'}) {
 	$opt->[OPTEXTRA]{'var'};
@@ -167,6 +171,13 @@ sub _varref {
 	eval q{\$} . $self->{'caller'} . q{::opt_} . $v; # XXX @, %
     }
 }
+# Formerly the varref method was private:
+sub _varref { shift->varref(@_) }
+
+sub optextra {
+    my($self, $opt, $arg) = @_;
+    $opt->[OPTEXTRA]{$arg};
+}
 
 sub _is_separator {
     my $opt = shift;
@@ -179,13 +190,13 @@ sub set_defaults {
     my $opt;
     foreach $opt ($self->_opt_array) {
 	if (defined $opt->[DEFVAL]) {
-	    my $ref = ref $self->_varref($opt);
+	    my $ref = ref $self->varref($opt);
 	    if      ($ref eq 'ARRAY') {
-		@ {$self->_varref($opt)} = @{ $opt->[DEFVAL] };
+		@ {$self->varref($opt)} = @{ $opt->[DEFVAL] };
 	    } elsif ($ref eq 'HASH') {
-		% {$self->_varref($opt)} = %{ $opt->[DEFVAL] };
+		% {$self->varref($opt)} = %{ $opt->[DEFVAL] };
 	    } elsif ($ref eq 'SCALAR') {
-		$ {$self->_varref($opt)} = $opt->[DEFVAL];
+		$ {$self->varref($opt)} = $opt->[DEFVAL];
 	    } else {
 		die "Invalid reference type for option $opt->[OPTNAME] while setting the default value (maybe you should specify <undef> as the default value)";
 	    }
@@ -216,16 +227,16 @@ sub load_options {
     my $opt;
     foreach $opt ($self->_opt_array) {
 	if (exists $loadoptions->{$opt->[OPTNAME]}) {
-	    if (ref $self->_varref($opt) eq 'CODE') {
-		&{$self->_varref($opt)} if $loadoptions->{$opt->[OPTNAME]};
-	    } elsif (ref $self->_varref($opt) eq 'ARRAY' &&
+	    if (ref $self->varref($opt) eq 'CODE') {
+		$self->varref($opt)->($opt, $loadoptions->{$opt->[OPTNAME]}) if $loadoptions->{$opt->[OPTNAME]};
+	    } elsif (ref $self->varref($opt) eq 'ARRAY' &&
 		     ref $loadoptions->{$opt->[OPTNAME]} eq 'ARRAY') {
-		@{ $self->_varref($opt) } = @{ $loadoptions->{$opt->[OPTNAME]} };
-	    } elsif (ref $self->_varref($opt) eq 'HASH' &&
+		@{ $self->varref($opt) } = @{ $loadoptions->{$opt->[OPTNAME]} };
+	    } elsif (ref $self->varref($opt) eq 'HASH' &&
 		     ref $loadoptions->{$opt->[OPTNAME]} eq 'HASH') {
-		%{ $self->_varref($opt) } = %{ $loadoptions->{$opt->[OPTNAME]} };
+		%{ $self->varref($opt) } = %{ $loadoptions->{$opt->[OPTNAME]} };
 	    } else {
-		$ {$self->_varref($opt)} = $loadoptions->{$opt->[OPTNAME]};
+		$ {$self->varref($opt)} = $loadoptions->{$opt->[OPTNAME]};
 	    }
 	}
     }
@@ -239,7 +250,7 @@ sub save_options {
     eval "require Data::Dumper";
     if ($@) {
 	warn $@;
-	die "No Data::Dumper\n";
+	$self->my_die("No Data::Dumper available, cannot save options.\n");
     } else {
 	if (open(OPT, ">$filename")) {
 	    my %saveoptions;
@@ -250,7 +261,7 @@ sub save_options {
 		    if ($opt->[OPTEXTRA]{'savevar'}) {
 			$ref = $opt->[OPTEXTRA]{'savevar'};
 		    } else {
-			$ref = $self->_varref($opt);
+			$ref = $self->varref($opt);
 		    }
 		    if (ref($ref) eq 'SCALAR') {
 			$saveoptions{$opt->[OPTNAME]} = $$ref;
@@ -259,6 +270,8 @@ sub save_options {
 		    }
 		}
 	    }
+	    local $Data::Dumper::Sortkeys = $Data::Dumper::Sortkeys = 1;
+	    local $Data::Dumper::Indent = $Data::Dumper::Indent = 1;
 	    if (Data::Dumper->can('Dumpxs')) {
 		# use faster version of Dump
 		print OPT
@@ -271,8 +284,7 @@ sub save_options {
 	    warn "Options written to $filename" if $DEBUG;
 	    1;
 	} else {
-	    warn "Can't write to $filename";
-	    die "Writing failed\n";
+	    $self->my_die("Writing to config file <$filename> failed: $!\n");
 	    undef;
 	}
     }
@@ -284,11 +296,11 @@ sub get_options {
     my $opt;
     foreach $opt ($self->_opt_array) {
 	$getopt{_getopt_long_string($opt->[OPTNAME], $opt->[OPTTYPE])} =
-	  $self->_varref($opt);
+	  $self->varref($opt);
 	# process aliases
 	foreach (@{$opt->[OPTEXTRA]{'alias'}}) {
 	    $getopt{_getopt_long_string($_, $opt->[OPTTYPE])} =
-	      $self->_varref($opt);
+	      $self->varref($opt);
 	}
     }
     require Getopt::Long;
@@ -370,16 +382,16 @@ sub usage {
 
 sub process_options {
     my($self, $former, $fromgui) = @_;
-    my $options = $self->{'options'};
-    foreach ($self->_opt_array) {
-	my $opt = $_->[OPTNAME];
+    my $bag = {};
+    foreach my $optdef ($self->_opt_array) {
+	my $opt = $optdef->[OPTNAME];
 
 	my $callback;
 	if ($fromgui) {
-	    $callback = $_->[OPTEXTRA]{'callback-interactive'};
+	    $callback = $optdef->[OPTEXTRA]{'callback-interactive'};
 	}
 	if (!$callback) {
-	    $callback = $_->[OPTEXTRA]{'callback'};
+	    $callback = $optdef->[OPTEXTRA]{'callback'};
 	}
 	if ($callback) {
 	    # no warnings here ... it would be too complicated to catch
@@ -389,25 +401,33 @@ sub process_options {
 	    # execute callback if value has changed
 	    if (!(defined $former
 		  && (!exists $former->{$opt}
-		      || $ {$self->_varref($_)} eq $former->{$opt}))) {
+		      || $ {$self->varref($optdef)} eq $former->{$opt}))) {
 		local($^W) = $old_w; # fall back to original value
-		&$callback;
+		&$callback(optdef => $optdef, bag => $bag);
 	    }
 	}
-	if ($_->[OPTEXTRA]{'strict'} &&
-	    UNIVERSAL::isa($_->[OPTEXTRA]{'choices'},'ARRAY')) {
+	if ($optdef->[OPTEXTRA]{'strict'} &&
+	    UNIVERSAL::isa($optdef->[OPTEXTRA]{'choices'},'ARRAY')) {
 	    # check for valid values (valid are: choices and default value)
-	    my $v = $ {$self->_varref($_)};
-	    my @choices = @{$_->[OPTEXTRA]{'choices'}};
-	    push(@choices, $_->[DEFVAL]) if defined $_->[DEFVAL];
-	    if (!grep($_ eq $v, @choices)) {
+	    my $v = $ {$self->varref($optdef)};
+	    my @choices = @{$optdef->[OPTEXTRA]{'choices'}};
+	    push(@choices, $optdef->[DEFVAL]) if defined $optdef->[DEFVAL];
+	    my $seen;
+	    for my $choice (@choices) {
+		my $value = (ref $choice eq 'ARRAY' ? $choice->[1] : $choice);
+		if ($value eq $v) {
+		    $seen = 1;
+		    last;
+		}
+	    }
+	    if (!$seen) {
 		if (defined $former) {
-		    warn "Not allowed: " . $ {$self->_varref($_)}
-		    . " for -$opt. Using old value $former->{$opt}";
-		    $ {$self->_varref($_)} = $former->{$opt};
+		    warn "Not allowed: " . $ {$self->varref($optdef)}
+		       . " for -$opt. Using old value $former->{$opt}";
+		    $ {$self->varref($optdef)} = $former->{$opt};
 		} else {
 		    die "Not allowed: "
-		      . $ {$self->_varref($_)} . " for -$opt\n"
+		      . $ {$self->varref($optdef)} . " for -$opt\n"
 		      . "Allowed is only: " . join(", ", @choices);
 		}
 	    }
@@ -415,6 +435,42 @@ sub process_options {
     }
 }
 
+sub my_die {
+    my($self, $msg, $is_safe) = @_;
+    my $use_tk;
+    if ($self->{'useerrordialog'} && defined &Tk::MainWindow::Existing) {
+	for my $mw (Tk::MainWindow::Existing()) {
+	    if (Tk::Exists($mw)) {
+		$use_tk = $mw;
+		last;
+	    }
+	}
+	if ($use_tk && !defined $is_safe) {
+	    for(my $i=0; $i<100; $i++) {
+		my(undef,undef,undef,$subroutine) = caller($i);
+		last if !defined $subroutine;
+		if ($subroutine eq '(eval)') {
+		    $use_tk = 0;
+		    last;
+		}
+	    }
+	}
+    }
+    if ($use_tk) {
+	eval {
+	    $use_tk->messageBox(-icon    => "error",
+				-message => $msg,
+				-title   => "Error",
+			       );
+	};
+	if ($@) {
+	    Carp::croak($msg);
+	}
+    } else {
+	Carp::croak($msg);
+    }
+}
+
 # try to work around weird browse entry
 sub _fix_layout {
     my($self, $frame, $widget, %args) = @_;
@@ -433,7 +489,7 @@ sub _fix_layout {
 sub _boolean_widget {
     my($self, $frame, $opt) = @_;
     ($self->_fix_layout($frame, "Checkbutton",
-			-variable => $self->_varref($opt)))[0];
+			-variable => $self->varref($opt)))[0];
 }
 
 sub _boolean_checkmark_widget {
@@ -441,7 +497,7 @@ sub _boolean_checkmark_widget {
     my($self, $frame, $opt) = @_;
     _create_checkmarks($frame);
     ($self->_fix_layout($frame, "Checkbutton",
-			-variable => $self->_varref($opt),
+			-variable => $self->varref($opt),
 			-image => $CHECKMARK_OFF,
 			-selectimage => $CHECKMARK_ON,
 			-indicatoron => 0,
@@ -456,7 +512,7 @@ sub _number_widget {
 			-to => $opt->[OPTEXTRA]{'range'}[1],
 			-showvalue => 1,
 			-resolution => ($opt->[OPTTYPE] =~ /f/ ? 0 : 1),
-			-variable => $self->_varref($opt)
+			-variable => $self->varref($opt)
 		       ))[0];
 }
 
@@ -479,15 +535,25 @@ sub _float_widget {
 }
 
 sub _list_widget {
+    my($self, $frame, $opt) = @_;
+    if ($opt->[OPTEXTRA]{'strict'} && grep { ref $_ eq 'ARRAY' } @{$opt->[OPTEXTRA]{'choices'}}) {
+	$self->_optionmenu_widget($frame, $opt);
+    } else {
+	$self->_browseentry_widget($frame, $opt);
+    }
+}
+
+sub _browseentry_widget {
     my($self, $frame, $opt) = @_;
     require Tk::BrowseEntry;
-    my %args = (-variable => $self->_varref($opt));
+    my %args = (-variable => $self->varref($opt));
     if ($opt->[OPTEXTRA]{'strict'}) {
 	$args{-state} = "readonly";
     }
     my $w = $frame->BrowseEntry(%args);
+    my %mapping;
     my @optlist = @{$opt->[OPTEXTRA]{'choices'}};
-    unshift(@optlist, $opt->[DEFVAL]) if defined $opt->[DEFVAL];
+    unshift @optlist, $opt->[DEFVAL] if defined $opt->[DEFVAL];
     my $o;
     my %seen;
     foreach $o (@optlist) {
@@ -499,6 +565,28 @@ sub _list_widget {
     $w;
 }
 
+sub _optionmenu_widget {
+    my($self, $frame, $opt) = @_;
+    require Tk::Optionmenu;
+    my $varref = $self->varref($opt);
+    # Have to remember value, otherwise Optionmenu would overwrite it...
+    my $value = $$varref;
+    my %args = (-variable => $varref,
+		-options => $opt->[OPTEXTRA]{'choices'},
+	       );
+    my $w = $frame->Optionmenu(%args);
+    if (defined $value) {
+	my $label = $value;
+	for my $choice (@{ $opt->[OPTEXTRA]{'choices'} }) {
+	    if (ref $choice eq 'ARRAY' && $choice->[1] eq $value) {
+		$label = $choice->[0];
+	    }
+	}
+	$w->setOption($label, $value);
+    }
+    $w;
+}
+
 sub _string_widget {
     my($self, $frame, $opt, %args) = @_;
     if (exists $opt->[OPTEXTRA]{'choices'}) {
@@ -508,7 +596,7 @@ sub _string_widget {
 	    ($frame, "Entry",
 	     (defined $opt->[OPTEXTRA]{'length'}
 	      ? (-width => $opt->[OPTEXTRA]{'length'}) : ()),
-	     -textvariable => $self->_varref($opt));
+	     -textvariable => $self->varref($opt));
 	if ($args{-restrict} || defined $opt->[OPTEXTRA]{'maxsize'}) {
 	    my $restrict_int   = sub { $_[0] =~ /^([+-]?\d+|)$/ };
 	    my $restrict_float = sub {
@@ -537,6 +625,10 @@ sub _string_widget {
 sub _dir_select {
     my($top, $curr_dir) = @_;
 
+    if ($top->can("chooseDirectory")) {
+	return $top->chooseDirectory(-initialdir => $curr_dir);
+    }
+
     if (eval { require Tk::DirSelect; Tk::DirSelect->VERSION("1.03"); 1 }) {
 	return $top->DirSelect(-directory => $curr_dir)->Show;
     }
@@ -596,7 +688,7 @@ sub _filedialog_widget {
     my $e;
     if (exists $opt->[OPTEXTRA]{'choices'}) {
 	require Tk::BrowseEntry;
-	$e = $topframe->BrowseEntry(-variable => $self->_varref($opt));
+	$e = $topframe->BrowseEntry(-variable => $self->varref($opt));
 	my @optlist = @{$opt->[OPTEXTRA]{'choices'}};
 	unshift(@optlist, $opt->[DEFVAL]) if defined $opt->[DEFVAL];
 	my $o;
@@ -608,13 +700,13 @@ sub _filedialog_widget {
                use Tk::PathEntry;
                my $real_e;
                ($e, $real_e) = $self->_fix_layout($topframe, "PathEntry",
-                                                  -textvariable => $self->_varref($opt));
+                                                  -textvariable => $self->varref($opt));
                # XXX Escape is already used for cancelling Tk::Getopt
                $real_e->bind("<$_>" => sub { $real_e->Finish }) for (qw/Return/);
                1;
            ') {
 	    ($e) = $self->_fix_layout($topframe, "Entry",
-				      -textvariable => $self->_varref($opt));
+				      -textvariable => $self->varref($opt));
 	}
     }
     $e->pack(-side => 'left');
@@ -653,7 +745,7 @@ sub _filedialog_widget {
 	       }
 	   }
 	   my($dir, $base, $file);
-	   my $act_val = $ {$self->_varref($opt)};
+	   my $act_val = $ {$self->varref($opt)};
 	   if ($act_val) {
 	       $dir  = File::Basename::dirname($act_val);
 	       $base = File::Basename::basename($act_val);
@@ -695,7 +787,7 @@ sub _filedialog_widget {
 	       }
 	   }
 	   if (defined $file && $file ne "") {
-	       $ {$self->_varref($opt)} = $file;
+	       $ {$self->varref($opt)} = $file;
 	   }
        });
     $b->pack(-side => 'left');
@@ -709,7 +801,7 @@ sub _geometry_widget {
 	 'Entry',
 	 (defined $opt->[OPTEXTRA]{'length'}
 	  ? (-width => $opt->[OPTEXTRA]{'length'}) : ()),
-	 -textvariable => $self->_varref($opt));
+	 -textvariable => $self->varref($opt));
     $topframe->Button(_get_curr_geometry_args($topframe),
 		      -command => sub {
 			  my $mw = $frame->MainWindow;
@@ -730,7 +822,7 @@ sub _color_widget {
 	 'Entry',
 	 (defined $opt->[OPTEXTRA]{'length'}
 	  ? (-width => $opt->[OPTEXTRA]{'length'}) : ()),
-	 -textvariable => $self->_varref($opt));
+	 -textvariable => $self->varref($opt));
     if ($frame->can("chooseColor")) {
 	$topframe->Button(-text => "...",
 			  -padx => 0, -pady => 0,
@@ -753,7 +845,7 @@ sub _font_widget {
 	 'Entry',
 	 (defined $opt->[OPTEXTRA]{'length'}
 	  ? (-width => $opt->[OPTEXTRA]{'length'}) : ()),
-	 -textvariable => $self->_varref($opt));
+	 -textvariable => $self->varref($opt));
     if (eval {require Tk::Font; require Tk::FontDialog; 1}) {
 	$topframe->Button(-text => "...",
 			  -padx => 0, -pady => 0,
@@ -872,7 +964,7 @@ sub _create_page {
 			   my $t = $f->Toplevel
 			       (-title => $self->{_string}{"helpfor"}
 				. " $label");
-			   $t->Label(-text => $opt->[OPTEXTRA]{'longhelp'},
+			   $t->Message(-text => $opt->[OPTEXTRA]{'longhelp'},
 				     -justify => 'left')->pack;
 			   $t->Button(-text => 'OK',
 				      -command => sub { $t->destroy }
@@ -891,18 +983,18 @@ sub _do_undo {
     foreach $opt ($self->_opt_array) {
 	next if $opt->[OPTEXTRA]{'nogui'};
 	if (exists $undo_options->{$opt->[OPTNAME]}) {
-	    my $ref = ref $self->_varref($opt);
+	    my $ref = ref $self->varref($opt);
 	    if      ($ref eq 'ARRAY') {
-		my @swap = @ {$self->_varref($opt)};
-		@ {$self->_varref($opt)} = @{ $undo_options->{$opt->[OPTNAME]} };
+		my @swap = @ {$self->varref($opt)};
+		@ {$self->varref($opt)} = @{ $undo_options->{$opt->[OPTNAME]} };
 		@{ $undo_options->{$opt->[OPTNAME]}} = @swap;
 	    } elsif ($ref eq 'HASH') {
-		my %swap = % {$self->_varref($opt)};
-		% {$self->_varref($opt)} = %{ $undo_options->{$opt->[OPTNAME]} };
+		my %swap = % {$self->varref($opt)};
+		% {$self->varref($opt)} = %{ $undo_options->{$opt->[OPTNAME]} };
 		%{ $undo_options->{$opt->[OPTNAME]}} = %swap;
 	    } elsif ($ref eq 'SCALAR') {
-		my $swap = $ {$self->_varref($opt)};
-		$ {$self->_varref($opt)} = $undo_options->{$opt->[OPTNAME]};
+		my $swap = $ {$self->varref($opt)};
+		$ {$self->varref($opt)} = $undo_options->{$opt->[OPTNAME]};
 		$undo_options->{$opt->[OPTNAME]} = $swap;
 	    } else {
 		die "Invalid reference type for option $opt->[OPTNAME]";
@@ -911,6 +1003,15 @@ sub _do_undo {
     }
 }
 
+sub option_dialog {
+    my($self, $top, %a) = @_;
+    my $button_pressed;
+    $a{'-buttonpressed'} = \$button_pressed;
+    $a{'-wait'} = 1;
+    $self->option_editor($top, %a);
+    $button_pressed;
+}
+
 sub option_editor {
     my($self, $top, %a) = @_;
     my $callback  = delete $a{'-callback'};
@@ -921,37 +1022,56 @@ sub option_editor {
     my $transient = delete $a{'-transient'};
     my $use_statusbar = delete $a{'-statusbar'};
     my $wait      = delete $a{'-wait'};
-    my $string    = delete $a{'-string'};
+    my $string    = delete $a{'-string'} || {};
     my $delay_page_create = (exists $a{'-delaypagecreate'}
 			     ? delete $a{'-delaypagecreate'}
 			     : 1);
-    if (!defined $string) {
-	$string = {'optedit'    => 'Option editor',
-		   'undo'       => 'Undo',
-		   'lastsaved'  => 'Last saved',
-		   'save'       => 'Save',
-		   'defaults'   => 'Defaults',
-		   'ok'         => 'OK',
-		   'apply'      => 'Apply',
-		   'cancel'     => 'Cancel',
-		   'helpfor'    => 'Help for:',
-		   'oksave'     => 'OK',
-	          };
+    my $page      = delete $a{'-page'};
+    my $button_pressed;
+    if (exists $a{'-buttonpressed'}) {
+	if (ref $a{'-buttonpressed'} ne "SCALAR") {
+	    die "The value for the -buttonpressed option has to be a SCALAR reference, not a " . ref($a{'-buttonpressed'}) . "\n";
+	}
+	$button_pressed = delete $a{'-buttonpressed'};
+    } else {
+	# dummy
+	$button_pressed = \do { my $dummy };
+    }
+    {
+	my %defaults = ('optedit'    => 'Option editor',
+			'undo'       => 'Undo',
+			'lastsaved'  => 'Last saved',
+			'save'       => 'Save',
+			'defaults'   => 'Defaults',
+			'ok'         => 'OK',
+			'apply'      => 'Apply',
+			'cancel'     => 'Cancel',
+			'helpfor'    => 'Help for:',
+			'oksave'     => 'OK',
+		       );
+	for my $key (keys %defaults) {
+	    next if exists $string->{$key};
+	    $string->{$key} = $defaults{$key};
+	}
     }
     $self->{_string} = $string;
 
+    if (defined $page) {
+	$self->{'raised'} = $page;
+    }
+
     # store old values for undo
     my %undo_options;
     my $opt;
     foreach $opt ($self->_opt_array) {
 	next if $opt->[OPTEXTRA]{'nogui'};
-	my $ref = ref $self->_varref($opt);
+	my $ref = ref $self->varref($opt);
 	if      ($ref eq 'ARRAY') {
-	    @{ $undo_options{$opt->[OPTNAME]} } = @ {$self->_varref($opt)};
+	    @{ $undo_options{$opt->[OPTNAME]} } = @ {$self->varref($opt)};
 	} elsif ($ref eq 'HASH') {
-	    %{ $undo_options{$opt->[OPTNAME]} } = % {$self->_varref($opt)};
+	    %{ $undo_options{$opt->[OPTNAME]} } = % {$self->varref($opt)};
 	} elsif ($ref eq 'SCALAR') {
-	    $undo_options{$opt->[OPTNAME]}      = $ {$self->_varref($opt)};
+	    $undo_options{$opt->[OPTNAME]}      = $ {$self->varref($opt)};
 	} else {
 	    die "Invalid reference type for option $opt->[OPTNAME]";
 	}
@@ -980,9 +1100,12 @@ sub option_editor {
     die "$@ while evaling $cmd" if $@;
     $opt_editor->transient($transient) if $transient;
     eval { $opt_editor->configure(-title => $string->{optedit}) };
+
     my $opt_notebook = ($dont_use_notebook ?
 			$opt_editor->Frame :
 			$opt_editor->NoteBook(-ipadx => 6, -ipady => 6));
+    $self->{Frame} = $opt_notebook;
+
     my($statusbar, $balloon);
     if (!$dont_use_balloon) {
 	if ($use_statusbar) {
@@ -1096,6 +1219,7 @@ sub option_editor {
                                  $self->{'raised'} = $opt_notebook->raised();
                              }
                              $opt_editor->destroy;
+			     $$button_pressed = 'ok';
                          }
                         );
         push @tiler_b, $ok_button;
@@ -1117,9 +1241,10 @@ sub option_editor {
 			     my $err = $@;
 			     $top->Unbusy;
 			     if ($err) {
-				 die $err;
+				 $self->my_die($err, 'safe');
 			     }
                              $opt_editor->destroy;
+			     $$button_pressed = 'ok';
                          }
                         );
         push @tiler_b, $ok_button;
@@ -1145,6 +1270,7 @@ sub option_editor {
 				 $self->{'raised'} = $opt_notebook->raised();
 			     }
 			     $opt_editor->destroy;
+			     $$button_pressed = 'cancel';
 			 }
 			);
 	push @tiler_b, $cancel_button;
@@ -1205,7 +1331,7 @@ sub option_editor {
     &$callback($self, $opt_editor) if $callback;
 
     if (!$dont_use_notebook && defined $self->{'raised'}) {
-	$opt_notebook->raise($self->{'raised'});
+	$self->raise_page($self->{'raised'});
     }
 
     $opt_editor->bind('<Escape>' => sub { $cancel_button->invoke });
@@ -1276,6 +1402,13 @@ sub _get_curr_geometry_args {
     }
 }
 
+sub raise_page {
+    my($self, $page) = @_;
+    my $opt_notebook = $self->{Frame};
+    $page = lc $page; # always lowercase in NoteBook internals
+    $opt_notebook->raise($page);
+}
+
 1;
 
 __END__
@@ -1415,6 +1548,15 @@ options.
 If set to true, do not use a safe compartment when loading options
 (see B<load_options>).
 
+=item -useerrordialog
+
+If set to true, then use an error dialog in user-relevant error
+conditions. Otherwise, the error message is printed to STDERR. This
+only includes errors which may happen in normal operation, but not
+programming errors like specifying erroneous options. If no Tk context
+is available (i.e. there is no MainWindow), then the error message
+will also be printed to STDERR.
+
 =back
 
 =item B<set_defaults>
@@ -1492,6 +1634,9 @@ Pops the option editor up. The editor provides facilitied for editing
 options, undoing, restoring to their default valued and saving to the
 default options file.
 
+The option editor is non-modal. For a modal dialog, see below for the
+L</option_dialog> method.
+
 The first argument is the parent widget. Further optional arguments are
 passed as a hash:
 
@@ -1520,6 +1665,12 @@ least the OK and Cancel buttons. The default set looks like this:
 
     -buttons => [qw/ok apply cancel undo lastsaved save defaults/]
 
+A minimal set could look like (here OK means accept and save)
+
+    -buttons => [qw/oksave cancel/]
+
+(and using less buttons is recommended).
+
 =item -toplevel
 
 Use another widget class instead of B<Toplevel> for embedding the
@@ -1551,14 +1702,19 @@ Use an additional status bar for help messages.
 
 =item -string
 
-Change button labels and title. This argument should be a hash reference
-with following keys: C<optedit>, C<undo>, C<lastsaved>, C<save>, C<defaults>,
-C<ok>, C<cancel>, C<helpfor>.
+Change button labels and title. This argument should be a hash
+reference with all or a subset of the following keys: C<optedit>,
+C<undo>, C<lastsaved>, C<save>, C<defaults>, C<ok>, C<cancel>,
+C<helpfor>.
 
 =item -wait
 
 Do not return immediately, but rather wait for the user pressing OK or Cancel.
 
+=item -page
+
+Raise the named notebook page (if grouping is used, see below).
+
 =back
 
 Since the option editor uses the C<NoteBook> widget, options may be
@@ -1618,6 +1774,12 @@ B<FileDialog> if B<subtype> is set to B<file>.
 
 =back
 
+=item B<option_dialog(>I<widget>, [I<arguments ...>]B<)>
+
+This method works like L</option_editor>, but it shows the option
+editor as a modal dialog. Additionaly, the return value is either
+B<ok> or B<cancel> depending on how the user quits the editor.
+
 =back
 
 =head1 OPTTABLE ARGUMENTS
@@ -1634,7 +1796,21 @@ An array of aliases also accepted by F<Getopt::Long>.
 =item callback
 
 Call a subroutine every time the option changes (e.g. after pressing
-on Apply, Ok or after loading).
+on Apply, Ok or after loading). The callback will get a hash with the
+following keys as argument:
+
+=over
+
+=item optdef
+
+The opttable item definition for this option.
+
+=item bag
+
+A hash reference which is persistent for this L</process_options>
+call. This can be used to share state between multiple callbacks.
+
+=back
 
 =item callback-interactive
 
@@ -1657,6 +1833,19 @@ A long help string used by B<option_editor>.
 
 An array of additional choices for the option editor.
 
+If C<strict> is set to a true value, then the elements of choices may
+also contain array references. In this case the first value of the
+"sub" array references are the display labels and the second value
+the used value. This is similar to L<Tk::Optionmenu> (in fact, for
+displaying this option an Optionmenu is used).
+
+     choices => ["one", "two", "three"]
+
+     choices => [["english"  => "en"],
+		 ["deutsch"  => "de"],
+		 ["hrvatski" => "hr"]]
+
+
 =item range
 
 An array with the beginning and end of a range for an integer or float value.
@@ -1671,11 +1860,13 @@ match either the choices or the range.
 Valid subtypes are I<file>, I<savefile>, I<dir>, I<geometry>, I<font>
 and I<color>. These can be used with string options. For I<file> and
 I<savefile>, the GUI interface will pop up a file dialog, using
-C<getOpenFile> for the former and C<getSaveFile> for the latter. For
-I<dir>, the GUI interface will pop up dialog for selecting
-directories. If the I<geometry> subtype is specified, the user can set
-the current geometry of the main window. The I<color> subtype is not
-yet implemented.
+L<getOpenFile|Tk::getOpenFile> for the former and
+L<getSaveFile|Tk::getOpenFile/getSaveFile> for the latter. For I<dir>,
+the GUI interface will pop up a dialog for selecting directories
+(using either L<Tk::chooseDirectory>, L<Tk::DirSelect>, or a custom
+dialog built on top of L<Tk::DirTree>). If the I<geometry> subtype is
+specified, the user can set the current geometry of the main window.
+The I<color> subtype is not yet implemented.
 
 =item var
 
@@ -1697,10 +1888,33 @@ Restrict the maximum number of characters in entries.
 =item widget
 
 This should be a reference to a subroutine for creating an own widget.
-Folowing arguments will be passed to this subroutine:
-a reference to the F<Tk::Getopt> object, Frame object, options entry.
-The subroutine should create a widget in the frame (packing is not
-necessary!) and should return a reference to the created widget.
+Folowing arguments will be passed to this subroutine: a reference to
+the F<Tk::Getopt> object, Frame object, and the options entry. The
+options entry should be used to get the variable reference with the
+C<_varref> method. The subroutine should create a widget in the frame
+(packing is not necessary!) and should return a reference to the
+created widget.
+
+A sample with an opttable entry for a custom numeric entry using the
+CPAN module L<Tk::NumEntry>:
+
+   ['numentry', '=i', 50,
+    range => [0, 100],
+    widget => sub { numentry_widget(@_) },
+   ],
+
+And C<numentry_widget> defined as:
+
+    use Tk::NumEntry;
+    sub numentry_widget {
+        my($self, $frame, $opt) = @_;
+        my $v = $self->varref($opt);
+	$frame->NumEntry(-minvalue => $opt->[3]{range}[0],
+    			 -maxvalue => $opt->[3]{range}[1],
+			 -variable => $v,
+    			 -value => $$v,
+    		        );
+    }
 
 =back
 
@@ -1742,6 +1956,23 @@ Here is an example for using a complex opttable description:
           # longer help displayed in the GUI's help window
           ]);
 
+=head1 OPTION ENTRY METHODS
+
+These methods operate on option entries:
+
+=over
+
+=item B<varref(>I<optentry>B<)>
+
+Return the variable reference for this entry.
+
+=item B<optextra(>I<optentry>, I<optarg>B<)>
+
+Return the value for the specified I<optarg> argument. See the
+L<OPTTABLE ARGUMENTS> section above for a list of possible arguments.
+
+=back
+
 =head1 COMPATIBILITY
 
 The argument to -opttable can be converted to a C<Getopt::Long>
@@ -1824,6 +2055,15 @@ This manual is confusing. In fact, the whole module is confusing.
 Setting variables in the editor should not set immediately the real variables.
 This should be done only by Apply and Ok buttons.
 
+There's no -font option (you have to use tricks with the option db and
+a special Name for the option editor):
+
+    $top->optionAdd("*somename*font" => $font);
+    $opt->option_editor(Name => "somename", ...);
+
+There's no (easy) way to get a large option editor fit on small
+screens. Try -font, if it would exist, but see above.
+
 =head1 AUTHOR
 
 Slaven Rezic <slaven@rezic.de>
